@@ -40,6 +40,11 @@ int debugMode = 0;
 /* Clickmode: 0 - first finger; 1 - second finger; 2 - center */
 int clickmode = 2;
 
+/* Maximum amount of milliseconds between two scrolling steps before easing stops. */
+#define MAX_EASING_INTERVAL 200
+/* Maximum amount of milliseconds between the last two scrolling steps for easing to start. */
+#define MAX_EASING_START_INTERVAL 200
+
 /* Continuation mode -- when 1, two finger gesture is continued when one finger is released. */
 #define CONTINUATION 1
 
@@ -154,6 +159,37 @@ int currentCenterX, currentCenterY;
 /* Profile of current window, if activated */
 Profile* currentProfile = NULL;
 
+/* Variables for easing: */
+/* The thread used for easing */
+pthread_t easingThread;
+/* Set to stop easing thread (as calling pthread_cancel didn't work well) */
+int stopEasing = 0;
+/* 1 if an easing thread is currently running. */
+int easingThreadActive = 0;
+/* Mutex for starting/stopping the easing thread */
+pthread_mutex_t easingMutex = PTHREAD_MUTEX_INITIALIZER;  
+/* Condition variable for the easing thread to wait */
+int easingWakeup = 0;
+pthread_cond_t  easingWaitingCond = PTHREAD_COND_INITIALIZER;
+/* Start Interval between two easing steps */
+int easingInterval = 0;
+/* 1 if the last X scroll command was to the right, -1 if it was to the left. */
+int easingDirectionX = 0;
+/* 1 if the last Y scroll command was down, -1 if it was up. */
+int easingDirectionY = 0;
+/* The profile for the easing */
+Profile* easingProfile;
+/* Time last scroll command in X/Y action was sent. */
+Time lastScrollXTime;
+Time lastScrollYTime;
+/* Interval between the last two X/Y scroll commands. */
+int lastScrollXIntv;
+int lastScrollYIntv;
+/*  Last values of lastScrollXIntv/lastScrollYIntv. */
+int lastLastScrollXIntv;
+int lastLastScrollYIntv;
+
+
 /* Handle errors by, well, throwing them away. */
 int invalidWindowHandler(Display *dsp, XErrorEvent *err) {
 	return 0;
@@ -291,14 +327,16 @@ void executeAction(Action* action, int whatToDo) {
 /* Returns the active top-level window. A top-level window is one that has WM_CLASS set.
  * May also return None. */
 Window getCurrentWindow() {
-
+	//if(debugMode) printf("Called getCurrentWindow\n");
 	/* First get the window that has the input focus */
 	Window currentWindow;
 	int revert;
 	XGetInputFocus(display, &currentWindow, &revert);
 
-	if (currentWindow == None)
+	if (currentWindow == None) {
+		//if(debugMode) printf("Leave getCurrentWindow\n");
 		return currentWindow;
+	}
 
 	/* Now go through parent windows until we find one with WM_CLASS set. */
 	Window root, parent;
@@ -307,43 +345,61 @@ Window getCurrentWindow() {
 
 
 	XClassHint* classHint = XAllocClassHint();
+	if(classHint == NULL) {
+		if(debugMode) printf("Couldn't allocate class hint!!\n");
+		return None;
+	}
 
 	int i = 0;
 	while (1) {
+		//if(debugMode) printf("in Loop\n");
 		i++;
 		if(i >= 5) {
+			if(debugMode) printf("Too many iterations in getCurrentWindow\n");
 			XFree(classHint);
+			//if(debugMode) printf("Leave getCurrentWindow\n");
 			return None;
 		}
-		if (currentWindow == root) {
+		if (currentWindow == root || currentWindow == None) {
+			//if(debugMode) printf("Reached root!\n");
 			/* No top-level window available. Should never happen. */
 			XFree(classHint);
+			//if(debugMode) printf("Leave getCurrentWindow\n");
 			return currentWindow;
 		}
 
+		//if(debugMode) printf("Call XGetClassHint!\n");
 		if (XGetClassHint(display, currentWindow, classHint) == 0) {
+			//if(debugMode) printf("Has no Class!\n");
 			/* Has no WM_CLASS, thus no top-level window */
 			if(XQueryTree(display, currentWindow, &root, &parent, &childWindows,
 					&childCount)) {
+				//if(debugMode) printf("Queried tree!\n");
 
 				if (childWindows != NULL)
 					XFree(childWindows);
 
 				if(currentWindow == parent) {
+					if(debugMode) printf("parent same as current!\n");
 					/* something wrong */
 					XFree(classHint);
+					//if(debugMode) printf("Leave getCurrentWindow\n");
 					return currentWindow;
 				}
 				/* Continue with parent until we find WM_CLASS */
 				currentWindow = parent;
 			} else {
+				//if(debugMode) printf("Could not query tree!\n");
 				XFree(classHint);
+				//if(debugMode) printf("Leave getCurrentWindow\n");
 				return currentWindow;
 			}
 		} else {
+			//if(debugMode) printf("Clean up class name!\n");
 			if(classHint->res_class != NULL) XFree(classHint->res_class);
 			if(classHint->res_name != NULL) XFree(classHint->res_name);
 			XFree(classHint);
+			//if(debugMode) printf("Leave getCurrentWindow\n");
 			return currentWindow;
 		}
 	}
@@ -403,9 +459,19 @@ void processFingers() {
 
 	if (active == 0)
 		return;
+
+	if(fingersDown != 0 && fingersWereDown == 0) {
+		stopEasingThread();
+	}
+
 	if (TWO_FINGERS_DOWN) {
 		/* Second finger touched (and maybe first too) */
 
+		lastScrollXTime = fairlyCurrentTime;
+		lastScrollYTime = fairlyCurrentTime;
+		lastScrollXIntv = 0; lastScrollYIntv = 0;		
+		lastLastScrollXIntv = 0; lastLastScrollYIntv = 0;		
+
 		maxDist = 0;
 
 		/* Don't move until we know what gesture we're performing. */
@@ -419,6 +485,8 @@ void processFingers() {
 		if(debugMode) {
 			if(currentProfile->windowClass != NULL) {
 				printf("Use profile '%s'\n", currentProfile->windowClass);
+			} else {
+				printf("Use default profile.\n");
 			}
 		}
 
@@ -477,8 +545,8 @@ void processFingers() {
 	} else if (TWO_FINGERS_UP) {
 		/* Second finger (and maybe also first) released */
 
-		/* If there was a scroll gesture and we have a brace action, perform release. */
 		if (amPerformingGesture == GESTURE_SCROLL) {
+			/* If there was a scroll gesture and we have a brace action, perform release. */
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollBraceAction),
 						EXECUTEACTION_RELEASE);
@@ -486,6 +554,30 @@ void processFingers() {
 				executeAction(&(currentProfile->scrollBraceAction),
 						EXECUTEACTION_RELEASE);
 			}
+			/* Start easing (has no effect if we only have brace action */
+			easingProfile = currentProfile;
+			if(debugMode) printf("Start easing\n");
+
+			/* Compensate for scrolling gestures getting a little bit slower at the end */
+			if(lastLastScrollXIntv < lastScrollXIntv && lastLastScrollXIntv != 0) lastScrollXIntv = lastLastScrollXIntv;
+			if(lastLastScrollYIntv < lastScrollYIntv && lastLastScrollYIntv != 0) lastScrollYIntv = lastLastScrollYIntv;
+
+			/* Check if scrolling intervals are not too long. Also check if last scrolling on an axis is longer
+			   ago than twice its interval, which means the scrolling has been stopped or extremely slowed down since. */
+			if(lastScrollYIntv == 0 || fairlyCurrentTime - lastScrollYTime > lastScrollYIntv * 2 || lastScrollYIntv > MAX_EASING_START_INTERVAL) easingDirectionY = 0;
+			if(lastScrollXIntv == 0 || fairlyCurrentTime - lastScrollXTime > lastScrollXIntv * 2 || lastScrollXIntv > MAX_EASING_START_INTERVAL) easingDirectionX = 0;
+			if(easingDirectionX != 0 || easingDirectionY != 0) {
+				if(easingDirectionY == 0) {
+					easingInterval = lastScrollXIntv;
+				} else if(easingDirectionX == 0) {
+					easingInterval = lastScrollYIntv;
+				} else {
+					/* As we only support one interval, use larger one so scrolling doesn't get	
+					   faster on the slower axis. */
+					easingInterval = lastScrollXIntv > lastScrollYIntv ? lastScrollXIntv : lastScrollYIntv;
+				}
+				startEasingThread();
+			}
 		}
 		/* If we haven't performed a gesture and haven't moved too far, perform tap action. */
 		if ((amPerformingGesture == GESTURE_NONE || amPerformingGesture
@@ -649,6 +741,10 @@ int checkGesture() {
 			return 0;
 
 		if (hscrolledBy > hscrollStep) {
+			easingDirectionX = 1;
+			lastLastScrollXIntv = lastScrollXIntv;
+			lastScrollXIntv = fairlyCurrentTime - lastScrollXTime;
+			lastScrollXTime = fairlyCurrentTime;
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollRightAction),
 						EXECUTEACTION_BOTH);
@@ -659,8 +755,11 @@ int checkGesture() {
 
 			gestureStartCenterX = gestureStartCenterX + hscrollStep;
 			return 1;
-		}
-		if (hscrolledBy < -hscrollStep) {
+		} else if (hscrolledBy < -hscrollStep) {
+			lastLastScrollXIntv = lastScrollXIntv;
+			lastScrollXIntv = fairlyCurrentTime - lastScrollXTime;
+			lastScrollXTime = fairlyCurrentTime;
+			easingDirectionX = -1;
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollLeftAction),
 						EXECUTEACTION_BOTH);
@@ -673,6 +772,10 @@ int checkGesture() {
 			return 1;
 		}
 		if (vscrolledBy > vscrollStep) {
+			lastLastScrollYIntv = lastScrollYIntv;
+			lastScrollYIntv = fairlyCurrentTime - lastScrollYTime;
+			lastScrollYTime = fairlyCurrentTime;
+			easingDirectionY = 1;
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollDownAction),
 						EXECUTEACTION_BOTH);
@@ -683,8 +786,11 @@ int checkGesture() {
 
 			gestureStartCenterY = gestureStartCenterY + vscrollStep;
 			return 1;
-		}
-		if (vscrolledBy < -vscrollStep) {
+		} else if (vscrolledBy < -vscrollStep) {
+			lastLastScrollYIntv = lastScrollYIntv;
+			lastScrollYIntv = fairlyCurrentTime - lastScrollYTime;
+			lastScrollYTime = fairlyCurrentTime;
+			easingDirectionY = -1;
 			if (currentProfile->scrollInherit) {
 				executeAction(&(defaultProfile.scrollUpAction),
 						EXECUTEACTION_BOTH);
@@ -789,6 +895,7 @@ void enterBlacklistedWindow() {
 
 /* Returns whether the given window is blacklisted */
 int isWindowBlacklisted(Window w) {
+	if(w == None) return 0;
 	int i;
 
 	XClassHint* classHint = XAllocClassHint();
@@ -810,6 +917,8 @@ int isWindowBlacklisted(Window w) {
 
 		if(classHint->res_class != NULL) XFree(classHint->res_class);
 		if(classHint->res_name != NULL) XFree(classHint->res_name);
+	} else {
+		printf("Found window with id %i and no class.\n", (int) w);
 	}
 	XFree(classHint);
 	return 0;
@@ -821,6 +930,7 @@ int isWindowBlacklisted(Window w) {
 void windowMapped(Window w) {
 
 	if (isWindowBlacklisted(w)) {
+		if(debugMode) printf("It's blacklisted.\n");
 		/* register for window focus events */
 		XSelectInput(display, w, EnterWindowMask | LeaveWindowMask);
 		if (getCurrentWindow() == w) {
@@ -828,7 +938,9 @@ void windowMapped(Window w) {
 			enterBlacklistedWindow();
 		}
 	} else {
+		if(debugMode) printf("It's not blacklisted.\n");
 		if (getCurrentWindow() == w) {
+			if(debugMode) printf("It's the current one!\n");
 			/* It is current and not blacklisted, so we might have left a blacklisted window */
 			leaveWindow();
 		}
@@ -851,6 +963,94 @@ void setScreenSize(XRRScreenChangeNotifyEvent * evt) {
 	}
 }
 
+
+/* Thread responsible for easing */
+void* easingThreadFunction(void* arg) {
+	if(debugMode) printf("Easing Thread started\n");
+	int nextInterval = easingInterval;
+	int lastTime = 0;
+	while(1) {
+		usleep((nextInterval - lastTime) * 1000);
+
+		pthread_mutex_lock(&easingMutex);
+		if(stopEasing || nextInterval > MAX_EASING_INTERVAL) {
+			stopEasing = 0;
+			/* Wait until woken up */
+			if(debugMode) printf("Easing thread goes to sleep. zZzZzZzZ...\n");
+			while(!easingWakeup) {
+				pthread_cond_wait(&easingWaitingCond, &easingMutex);
+			}
+			easingWakeup = 0;
+			if(debugMode) printf("*rrrrring* Easing thread woken up!\n");
+			nextInterval = easingInterval;
+		}
+		pthread_mutex_unlock(&easingMutex);
+
+		if(debugMode) printf("Easing step\n");
+		if (easingProfile->scrollInherit) {
+			if(easingDirectionY == -1) {
+				executeAction(&(defaultProfile.scrollUpAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionY == 1) {
+				executeAction(&(defaultProfile.scrollDownAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionX == -1) {
+				executeAction(&(defaultProfile.scrollLeftAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionX == 1) {
+				executeAction(&(defaultProfile.scrollRightAction),
+					EXECUTEACTION_BOTH);
+			}
+		} else {
+			if(easingDirectionY == -1) {
+				executeAction(&(easingProfile->scrollUpAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionY == 1) {
+				executeAction(&(easingProfile->scrollDownAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionX == -1) {
+				executeAction(&(easingProfile->scrollLeftAction),
+					EXECUTEACTION_BOTH);
+			}
+			if(easingDirectionX == 1) {
+				executeAction(&(easingProfile->scrollRightAction),
+					EXECUTEACTION_BOTH);
+			}
+		}
+		nextInterval = (int) (((float) nextInterval) * 1.15);
+	}
+	return NULL;
+}
+
+/* Starts the easing; profile, interval and directions have to be set before. */
+void startEasingThread() {
+	pthread_mutex_lock(&easingMutex);
+	stopEasing = 0;
+	if(!easingThreadActive) {
+		pthread_create(&easingThread, NULL,
+					easingThreadFunction, NULL);
+		easingThreadActive = 1;
+	} else {
+		/* wake up easing thread */
+		easingWakeup = 1;
+		pthread_cond_broadcast(&easingWaitingCond);
+	}
+	pthread_mutex_unlock(&easingMutex);
+}
+/* Stops the easing. */
+void stopEasingThread() {
+	pthread_mutex_lock(&easingMutex);
+	if(easingThreadActive) {
+		stopEasing = 1;
+	}
+	pthread_mutex_unlock(&easingMutex);
+}
+
 /* X thread that processes X events in parallel to kernel device loop */
 void* xLoopThreadFunction(void* arg) {
 	XEvent ev;
@@ -1192,7 +1392,7 @@ int main(int argc, char **argv) {
 
 		XIFreeDeviceInfo(info);
 
-		printf("XInput device id is %i.\n", deviceID);
+		if(debugMode) printf("XInput device id is %i.\n", deviceID);
 
 		/* Prepare by reading calibration */
 		readCalibrationData(1);
@@ -1213,9 +1413,15 @@ int main(int argc, char **argv) {
 		XTestGrabControl(display, True);
 		active = 0;
 		checkRunningWindows();
-		/* Don't wait until release but activate right now */
-		if (activateAtRelease) {
-			activateAtRelease = 0;
+		Window w = getCurrentWindow();
+		activateAtRelease = 0;
+		if(debugMode) printf("Current Window: %i\n", (int) w);
+
+		if(isWindowBlacklisted(w)) {
+			/* Current is blacklisted, call enterBlacklistedWindow */
+			enterBlacklistedWindow();
+		} else {
+			/* Current is not blacklisted, so we might have left a blacklisted window */
 			activate();
 		}
 
