@@ -34,6 +34,9 @@ struct Action {
 	int modifier;
 };
 
+void startEasingThread();
+void stopEasingThread();
+
 #define MODIFIER_SHIFT 1
 #define MODIFIER_CONTROL 2
 #define MODIFIER_ALT 4
